"""
Module for running a pipeline of middleware components in an ML context.

This module provides classes and exceptions for building and running a pipeline where
batches from a data loader are processed sequentially through a series of middleware
components, which receive a shared state and batch (data) as their input.
"""

from collections.abc import MutableMapping, Sequence, Sized
from typing import Any

from .errors import AbortPipeline, SkipStep, ValidationError
from .loader import Loader
from .middleware import Middleware, ProcessingStep, StrMapping

__all__ = ["Pipeline", "PipelineWrapper"]


class Pipeline:
    """
    Runs sequentially a series of middleware components to process data.

    The Pipeline builds a chain of middlewares by wrapping each middleware in reverse
    order (last middleware wraps an empty processing step) so that they execute in
    the provided order.
    """

    def __init__(
        self,
        middlewares: Sequence[Middleware],  # type: ignore[type-arg]
        step_name: str = "step",
    ) -> None:
        """
        Initialize the pipeline.

        An empty sink callable is wrapped by each middleware component (in reverse
        order), so that they execute sequentially in the order provided when
        processing a batch from the data loader.

        Args:
            middlewares: A sequence of middleware components.
            step_name: The key used to record the current step index in the
                state. (Default: "step")

        """
        self.middlewares = middlewares
        self.step_name = step_name

        self._run = _empty_sink
        for mware in reversed(self.middlewares):
            self._run = mware.wrap(self._run)

    def validate(
        self,
        state_fields: set[str],
        data_fields: set[str],
        sized_data_loader: bool,
    ) -> None:
        """
        Perform a "dry run" validation of the pipeline to check that all required
        fields will be present in the state and data at the correct stages.

        This method verifies that each middleware's required fields are satisfied
        by performing two passes through the list of middlewares:

        1. **Forward Pass (pre=True)**:
            - Iterates over the middlewares in the order they are provided.
            - Each middleware checks whether the current `state_fields` and
              `data_fields` satisfy its requirements for the "pre" phase (input fields).
            - If a middleware contributes new fields to the state or data on the way in,
              those fields are added to the respective sets for subsequent middlewares.
              Typically, middleware should not change the state fields, with the
              exception of `PipelineWrapper` (which adds its own step/num_steps fields).

        2. **Reverse Pass (pre=False)**:
            - Iterates over the middlewares in reverse order.
            - Each middleware checks whether the (now possibly updated) `state_fields`
              and `data_fields` satisfy its requirements for the "post" phase.
            - If a middleware introduces any additional fields on the way out, those
              fields are also accumulated for the previous middlewares. Again,
              middlewares should not add or alter state fields here.

        This method always adds `step_name` to `state_fields`. If `sized_data_loader`
        is True, it also adds `num_{step_name}s`.

        Args:
            state_fields: The set of keys that the pipeline expects in the shared
               `state` object at the start of processing.
            data_fields: The set of keys that each data batch is expected to have
                when produced by the data loader.
            sized_data_loader: Whether or not the data loader has a length, which
                determines if `num_{step_name}s` will be added to the state.

        Raises:
            ValidationError: If any required fields are missing in either the state
                or the data at any stage of validation.

        """
        # Create copy of fields sets to avoid modifying the original
        data_fields = set(data_fields)
        state_fields = set(state_fields)

        state_fields.add(self.step_name)
        if sized_data_loader:
            state_fields.add(f"num_{self.step_name}s")

        for i, mware in enumerate(self.middlewares):
            try:
                mware._validate(state_fields, data_fields, pre=True)  # noqa: SLF001
            except ValidationError as e:
                msg = (
                    "Validation error by middleware"
                    f" {type(mware).__name__}, at index {i}."
                )
                raise ValidationError(msg) from e

        for i, mware in enumerate(reversed(self.middlewares)):
            try:
                mware._validate(state_fields, data_fields, pre=False)  # noqa: SLF001
            except ValidationError as e:
                ind = len(self.middlewares) - i - 1
                msg = (
                    "Validation error by middleware"
                    f" {type(mware).__name__}, at index {ind}."
                )
                raise ValidationError(msg) from e

    def run(
        self,
        state: MutableMapping[str, Any],
        data_loader: Loader,
        validate: bool = True,
    ) -> None:
        """
        Process batches from the data loader through the middleware chain.

        The pipeline processes batches sequentially. Before looping over the
        data loader, if `validate=True`, the pipeline calls `validate` to checks
        that the state and the data contain all fields required by the middlewares.

        The batch is then passed through the middleware chain for processing. The data
        passes through the middlewares sequentially, in the order in which they are
        given in the list. Each middleware "passes" the control on to the next one by
        calling `next_step` in its wrapped function. The final middleware does not need
        to call next; in case it does, it simply calls an empty processing step.

        Then, the flow passes backwards: from the last middleware to the first one, with
        execution in each middleware continuing after the call to `next_step`.

        Here is an illustration of this flow:

        ```
        (data, state)
            ▼
        ┌────────────────┐
        │ Middleware 1   │
        └────────────────┘
            ▼          ▲
        ┌────────────────┐
        │       ...      │
        └────────────────┘
            ▼          ▲
        ┌────────────────┐
        │ Middleware N   │
        └────────────────┘
        ```

        If a middleware raises an AbortPipeline exception, processing stops immediately.
        If a middleware raises a SkipStep exception, that batch is skipped and
        processing continues with the next batch.

        On each step, the current step index is recorded in the state using the key
        specified by `step_name`. If data loader has a length, this is added to state
        under the key `num_{step_name}s` - for example, if `step_name="step"`, this
        would be `num_steps`.

        Before the processing begins, the `on_start` hook of each middleware is
        called, and after all processing finishes (also in the case of `AbortPipeline`
        exception being raised), the `on_finish` hook of each middlewar is called. The
        shared state is provided to each call. Both of these calls will call middlewares
        in the order they are provided (unlike processing, where on the backwards pass
        the order is reversed). Also, the `on_start` call will happen **before** any
        validation takes place.

        Args:
            state: A dictionary representing the shared pipeline state.
            data_loader: An iterable that yields batches of data (as dictionaries).
            validate: Whether to validate the required fields before starting the
                main loop. (Default: True)

        """
        for mware in self.middlewares:
            mware.on_start(state)

        if isinstance(data_loader, Sized):
            state[f"num_{self.step_name}s"] = len(data_loader)

        if validate:
            self.validate(
                set(state.keys()),
                data_loader.data_fields,
                isinstance(data_loader, Sized),
            )

        for step, data in enumerate(data_loader):
            state[self.step_name] = step

            try:
                self._run(state=state, data=data)
            except AbortPipeline:
                break
            except SkipStep:
                continue

        for mware in self.middlewares:
            mware.on_finish(state)


def _empty_sink(state: StrMapping, data: StrMapping) -> None:  # noqa: ARG001
    """An empty processing step (sink)."""
    return


class PipelineWrapper(Middleware[StrMapping, StrMapping]):
    """
    A middleware that encapsulates an entire `Pipeline` as a single middleware step.

    The wrapped pipeline is run in the "pre" step (forward pass), before the outer
    pipeline calls `next_step`. This means that for each batch passed to this
    middleware, it will first execute the wrapped pipeline by calling
    `pipeline.run` on the provided `data_loader`, and then continue to the next
    component in the outer pipeline.
    """

    def __init__(self, pipeline: Pipeline, data_loader: Loader) -> None:
        """
        Initialize a `PipelineWrapper` instance.

        Args:
            pipeline: The pipeline to be run when this middleware is invoked.
            data_loader: The data loader consumed by the wrapped pipeline.

        """
        super().__init__()
        self.pipeline = pipeline
        self.data_loader = data_loader

    def _validate(
        self,
        state_fields: set[str],
        data_fields: set[str],  # noqa: ARG002
        pre: bool = True,
    ) -> None:
        """
        Validate that the wrapped pipeline has all the fields it needs.

        During the "pre" phase, this method calls the wrapped pipeline's
        validation to ensure that `state_fields` and `data_fields` are sufficient
        for its own middlewares. This is effectively a nested validation pass.

        Args:
            state_fields: Current set of state fields available in the outer pipeline.
            data_fields: Current set of data fields available in the outer pipeline.
            pre: Whether this validation pass is happening before or after the batch
                is processed. The pipeline's validation is only performed during the
                pre phase.

        Raises:
            ValidationError: If the wrapped pipeline determines that required fields
                are missing.

        """
        if pre:
            size_data_loader = isinstance(self.data_loader, Sized)
            self.pipeline.validate(
                state_fields, self.data_loader.data_fields, size_data_loader
            )

    def wrap(
        self, next_step: ProcessingStep[StrMapping, StrMapping]
    ) -> ProcessingStep[StrMapping, StrMapping]:
        """
        Wrap the next step in a function that runs the wrapped pipeline first.

        This method creates and returns a new processing function. When called, the
        function will:
          1. Execute the wrapped pipeline by calling `pipeline.run` on the provided
             `data_loader`, skipping validation (since it's handled separately).
          2. Call `next_step` to continue with the rest of the outer pipeline.

        Args:
            next_step: The subsequent processing step in the outer pipeline.

        Returns:
            A callable that, when invoked with `(state, data)`, runs the wrapped
            pipeline and then calls `next_step`.

        """

        def wrapped(state: StrMapping, data: StrMapping) -> None:
            self.pipeline.run(state, self.data_loader, validate=False)
            next_step(state, data)

        return wrapped



from typing import Any

import pytest

from middl import Middleware, ProcessingStep, ValidationError


class _TestMiddleware(Middleware[Any, Any]):
    def __init__(self) -> None:
        super().__init__()

    def wrap(self, next_step: ProcessingStep[Any, Any]) -> ProcessingStep[Any, Any]:
        def wrapped(state: Any, data: Any) -> None:
            next_step(state, data)

        return wrapped


def test_validate_success_pre() -> None:
    mw = _TestMiddleware()
    mw.requires_state_fields = {"state_req1"}
    mw.requires_data_fields_pre = {"data_req1"}
    mw.provides_data_fields_pre = {"data_prov1"}
    state_fields = {"state_req1", "some_other_state"}
    data_fields = {"data_req1", "some_other_data"}
    mw._validate(state_fields, data_fields, pre=True)
    assert "data_prov1" in data_fields


def test_validate_success_post() -> None:
    mw = _TestMiddleware()
    mw.requires_state_fields = {"state_req2"}
    mw.requires_data_fields_post = {"data_req2"}
    mw.provides_data_fields_post = {"data_prov2"}
    state_fields = {"state_req2", "another_state"}
    data_fields = {"data_req2", "another_data"}
    mw._validate(state_fields, data_fields, pre=False)
    assert "data_prov2" in data_fields


def test_validate_missing_state_fields_pre() -> None:
    mw = _TestMiddleware()
    mw.requires_state_fields = {"state_req1"}
    mw.requires_data_fields_pre = {"data_req1"}
    mw.provides_data_fields_pre = {"data_prov1"}
    state_fields = {"some_other_state"}
    data_fields = {"data_req1"}
    with pytest.raises(ValidationError, match="Missing state fields {'state_req1'}"):
        mw._validate(state_fields, data_fields, pre=True)


def test_validate_missing_state_fields_post() -> None:
    mw = _TestMiddleware()
    mw.requires_state_fields = {"state_req2"}
    mw.requires_data_fields_post = {"data_req2"}
    mw.provides_data_fields_post = {"data_prov2"}
    state_fields = {"some_other_state"}
    data_fields = {"data_req2"}
    with pytest.raises(ValidationError, match="Missing state fields {'state_req2'}"):
        mw._validate(state_fields, data_fields, pre=False)


def test_validate_missing_data_fields_pre() -> None:
    mw = _TestMiddleware()
    mw.requires_state_fields = {"state_req1"}
    mw.requires_data_fields_pre = {"data_req1", "data_req2"}
    mw.provides_data_fields_pre = {"data_prov1"}
    state_fields = {"state_req1"}
    data_fields = {"data_req1"}
    with pytest.raises(ValidationError, match="Missing data pre fields {'data_req2'}"):
        mw._validate(state_fields, data_fields, pre=True)


def test_validate_missing_data_fields_post() -> None:
    mw = _TestMiddleware()
    mw.requires_state_fields = {"state_req1"}
    mw.requires_data_fields_post = {"data_reqX"}
    mw.provides_data_fields_post = {"data_provX"}
    state_fields = {"state_req1"}
    data_fields = {"some_other_data"}
    with pytest.raises(ValidationError, match="Missing data post fields {'data_reqX'}"):
        mw._validate(state_fields, data_fields, pre=False)



"""Module defining middl-specific errors."""

__all__ = ["AbortPipeline", "SkipStep", "ValidationError"]


class SkipStep(Exception):  # noqa: N818
    """
    Exception to skip processing of the current batch.

    When raised by a middleware, the pipeline skips the current step and
    continues with the next one.
    """


class AbortPipeline(Exception):  # noqa: N818
    """
    Exception to abort the entire pipeline execution.

    When raised by a middleware, the pipeline immediately exits.
    """


class ValidationError(Exception):
    """
    Exception raised when pipeline validation fails due to missing required fields.

    This exception indicates that the state or the batch from the data loader is missing
    fields required by one of the middlewares.
    """



"""
Composable middleware components for creating machine learning pipelines.
"""

from .core import (
    AbortPipeline,
    EmptyLoader,
    Loader,
    Middleware,
    Pipeline,
    PipelineWrapper,
    ProcessingStep,
    SkipStep,
    StrMapping,
    ValidationError,
    WrappedSizedLoader,
    WrappedUnsizedLoader,
    wrap_iterable,
)

__all__ = [
    "AbortPipeline",
    "EmptyLoader",
    "Loader",
    "Middleware",
    "Pipeline",
    "PipelineWrapper",
    "ProcessingStep",
    "SkipStep",
    "StrMapping",
    "ValidationError",
    "WrappedSizedLoader",
    "WrappedUnsizedLoader",
    "wrap_iterable",
]


__version__ = "0.0.1a7"



"""
Module containing core objects for implementing pipelines and creating
concrete middleware subclasses.
"""

from .errors import AbortPipeline, SkipStep, ValidationError
from .loader import (
    EmptyLoader,
    Loader,
    WrappedSizedLoader,
    WrappedUnsizedLoader,
    wrap_iterable,
)
from .middleware import Middleware, ProcessingStep, StrMapping
from .pipeline import Pipeline, PipelineWrapper

__all__ = [
    "AbortPipeline",
    "EmptyLoader",
    "Loader",
    "Middleware",
    "Pipeline",
    "PipelineWrapper",
    "ProcessingStep",
    "SkipStep",
    "StrMapping",
    "ValidationError",
    "WrappedSizedLoader",
    "WrappedUnsizedLoader",
    "wrap_iterable",
]



"""
Module defining middleware protocols and base classes.

This module provides the ProcessingStep protocol for the processing step and
the Middleware abstract base class for middleware components.
"""

from abc import ABC, abstractmethod
from collections.abc import MutableMapping
from typing import Any, Generic, Protocol, TypeVar

from .errors import ValidationError

__all__ = ["Middleware", "ProcessingStep", "StrMapping"]


StrMapping = MutableMapping[str, Any]
StateType_contra = TypeVar("StateType_contra", bound=StrMapping, contravariant=True)
DataType_contra = TypeVar("DataType_contra", bound=StrMapping, contravariant=True)


class ProcessingStep(Protocol[StateType_contra, DataType_contra]):
    """
    Protocol for a (callable) processing step of the pipeline.

    Needs to be a callable that accepts a shared state and a data batch,
    both as dictionaries (mappings) with string keys. Implementations should
    use concrete types for the state and data to leverage static type checking.
    """

    def __call__(self, state: StateType_contra, data: DataType_contra) -> None:
        """Perform the processing step."""


class Middleware(ABC, Generic[StateType_contra, DataType_contra]):
    """
    Abstract base class for middleware components.

    This class is generic over two types: one for the state and one for the data.
    Concrete middleware implementations should specify these generic types (typically
    as concrete TypedDict types) to benefit from IDE autocompletion and static type
    checking.

    Concrete implementations need to override the `wrap` method. Optionally, they
    can also override `on_start` and `on_finish` callbacks, in case there is some
    initialization/teardown that needs to be done at the start and at the end of
    pipeline execution.

    Attributes:
        requires_state_fields: The state fields required by this middleware.
        requires_data_fields_pre: The data fields required before processing.
        provides_data_fields_pre: The data fields provided after pre-processing.
        requires_data_fields_post: The data fields required after processing.
        provides_data_fields_post: The data fields provided after post-processing.

    Example:
    ```python
        from typing import TypedDict

        from middle import Middleware, StrMapping

        class MyState(TypedDict):
            step: int

        class MyData(TypedDict):
            loss: int

        class ExampleMiddleware(Middleware[MyState, MyData]):
            def __init__(self) -> None:
                super().__init__()
                self.requires_state_fields = {"step"}
                self.requires_data_fields_post = {"loss"}
                self.provides_data_fields_pre = {"value"}

            def wrap(
                self, next_step: ProcessingStep[StrMapping, StrMapping],
            ) -> ProcessingStep[MyState, MyData]:
                def wrapped(state: MyState, data: MyData) -> None:
                    print(f"Step: {state['step']}")
                    data["value"] = 1
                    next_step(state, data)
                    print(f"Loss: {data['loss']}")
                return wrapped
    ```

    """

    requires_state_fields: set[str]

    requires_data_fields_pre: set[str]
    provides_data_fields_pre: set[str]
    requires_data_fields_post: set[str]
    provides_data_fields_post: set[str]

    def __init__(self) -> None:
        """
        Initialize the middleware with empty sets for required and provided fields.
        """
        self.requires_state_fields = set()

        self.requires_data_fields_pre = set()
        self.provides_data_fields_pre = set()
        self.requires_data_fields_post = set()
        self.provides_data_fields_post = set()

    @abstractmethod
    def wrap(
        self,
        next_step: ProcessingStep[StrMapping, StrMapping],
    ) -> ProcessingStep[StateType_contra, DataType_contra]:
        """
        Wrap the next processing step in the middleware chain with this
        middleware's processing logic.

        Concrete implementations should return a callable that applies
        this middleware's behavior before and/or after invoking the next callable.

        Args:
            next_step: The next callable in the processing chain. This callable accepts
                a state and data and returns None.

        Returns:
            A callable that wraps the next_step with this middleware's behavior.

        Example:
            A simple middleware implementation that prints messages before and after
            invoking the next step:

                def wrap(self, next_step):
                    def wrapped(state, data):
                        print("Before calling next step")
                        next_step(state, data)
                        print("After calling next step")
                    return wrapped

        """
        raise NotImplementedError

    def on_start(self, state: StateType_contra) -> None:
        """
        Callback invoked at the start of the pipeline execution.

        This method is called once before any data is processed by the pipeline.
        Middleware implementations can override this method to any tasks needed
        before the processing begins - for example, a progress bar middleware
        would create (display on screen) a progress bar at this point.

        Args:
            state: A mapping representing the shared state of the pipeline.

        """

    def on_finish(self, state: StateType_contra) -> None:
        """
        Callback invoked at the end of the pipeline execution.

        This method is called once after all data batches have been processed or when
        the pipeline terminates prematurely. Middleware implementations can override
        this method to perform any necessary finalization tasks - for example, a
        progress bar middleware would remove a progress bar at this point.

        Args:
            state: A mapping representing the shared state of the pipeline.

        """

    def _validate(
        self, state_fields: set[str], data_fields: set[str], pre: bool = True
    ) -> None:
        """
        Validate the fields of the middleware, given input fields.

        This method checks that all the fields required by the middleware are
        present in the input fields.

        Arguments:
            state_fields: Fields of the input state mapping.
            data_fields: Fields of the input data mapping.
            pre: If `True`, perform the validation for the forward pass (pre),
                if `False`, perform the validation for the backward pass (post).

        """
        if not self.requires_state_fields.issubset(state_fields):
            msg = (
                "Missing state fields"
                f" {self.requires_state_fields.difference(state_fields)}"
            )
            raise ValidationError(msg)

        if pre:
            if not self.requires_data_fields_pre.issubset(data_fields):
                msg = (
                    "Missing data pre fields"
                    f" {self.requires_data_fields_pre.difference(data_fields)}"
                )
                raise ValidationError(msg)

            data_fields.update(self.provides_data_fields_pre)
        else:
            if not self.requires_data_fields_post.issubset(data_fields):
                msg = (
                    "Missing data post fields"
                    f" {self.requires_data_fields_post.difference(data_fields)}"
                )
                raise ValidationError(msg)

            data_fields.update(self.provides_data_fields_post)



from collections.abc import Generator

import pytest

from middl import EmptyLoader, WrappedSizedLoader, WrappedUnsizedLoader, wrap_iterable


def test_empty_loader() -> None:
    length = 3
    gen = EmptyLoader(length)

    assert len(gen) == length

    output_1 = list(gen)

    # modify one item to check that new dictionaries are generated
    output_1[0] = {1: 1}  # type: ignore[dict-item]
    output_2 = list(gen)

    assert output_1 == [{1: 1}, {}, {}]  # type: ignore[comparison-overlap]
    assert output_2 == [{}, {}, {}]


def test_zero_fields() -> None:
    data = [(), ()]
    data_fields: tuple[str, ...] = ()
    wrapped = wrap_iterable(data, data_fields)
    results = list(wrapped)
    assert len(results) == len(data)
    for r in results:
        assert r == {}


def test_one_field() -> None:
    data = [("val1",), ("val2",)]
    data_fields = ("field",)
    wrapped = wrap_iterable(data, data_fields)
    results = list(wrapped)
    assert len(results) == len(data)
    assert results[0] == {"field": "val1"}
    assert results[1] == {"field": "val2"}


def test_three_fields() -> None:
    data = [
        ("a1", "b1", "c1"),
        ("a2", "b2", "c2"),
    ]
    data_fields = ("A", "B", "C")
    wrapped = wrap_iterable(data, data_fields)
    results = list(wrapped)
    assert len(results) == len(data)
    for r in results:
        assert set(r.keys()) == set(data_fields)
        assert len(r.values()) == len(data[0])
        for k, v in r.items():
            # Make sure keys are matched to the item at its
            # corresponding position.
            assert k[0] == v[0].upper()


def test_batch_size_mismatch() -> None:
    data = [("x", "y")]
    data_fields = ("A", "B", "C")
    with pytest.raises(ValueError, match="zip\\(\\) argument"):
        list(wrap_iterable(data, data_fields))


def test_sized_loader_length() -> None:
    data = [
        ("item1",),
        ("item2",),
        ("item3",),
    ]
    data_fields = ("only_key",)
    wrapped = wrap_iterable(data, data_fields)
    assert isinstance(wrapped, WrappedSizedLoader)
    assert len(wrapped) == len(data)


def test_wrap_iterable_sized() -> None:
    data = [("one",), ("two",)]
    data_fields = ("key",)
    wrapped = wrap_iterable(data, data_fields)
    assert isinstance(wrapped, WrappedSizedLoader)


def test_wrap_iterable_unsized() -> None:
    def gen() -> Generator[tuple[str], None, None]:
        yield ("first",)
        yield ("second",)

    data_fields = ("only_key",)
    wrapped = wrap_iterable(gen(), data_fields)
    assert isinstance(wrapped, WrappedUnsizedLoader)



from typing import Any
from unittest.mock import MagicMock

import pytest

from middl import (
    AbortPipeline,
    EmptyLoader,
    Middleware,
    Pipeline,
    PipelineWrapper,
    ProcessingStep,
    SkipStep,
    ValidationError,
    WrappedUnsizedLoader,
    wrap_iterable,
)


class AccMiddleware(Middleware[Any, Any]):
    def wrap(self, next_step: ProcessingStep[Any, Any]) -> ProcessingStep[Any, Any]:
        def wrapped(state: Any, data: Any) -> None:
            next_step(state, data)
            state["acc"].append(state["value"])

        return wrapped


class _SimpleMiddleware(Middleware[Any, Any]):
    def wrap(self, next_step: ProcessingStep[Any, Any]) -> ProcessingStep[Any, Any]:
        def wrapped(state: Any, data: Any) -> None:
            return next_step(state, data)

        return wrapped


def test_validate_missing_pre_field() -> None:
    sm = _SimpleMiddleware()
    sm.requires_data_fields_pre = {"miss"}

    pipe = Pipeline(middlewares=[sm])

    with pytest.raises(
        ValidationError,
        match="Validation error by middleware _SimpleMiddleware, at index 0.",
    ):
        pipe.validate(set(), set(), sized_data_loader=False)


def test_validate_missing_post_field() -> None:
    sm = _SimpleMiddleware()
    sm.requires_data_fields_post = {"miss"}

    pipe = Pipeline(middlewares=[sm])

    with pytest.raises(
        ValidationError,
        match="Validation error by middleware _SimpleMiddleware, at index 0.",
    ):
        pipe.validate(set(), set(), sized_data_loader=False)


def test_validate_missing_state_field() -> None:
    sm = _SimpleMiddleware()
    sm.requires_state_fields = {"miss"}

    pipe = Pipeline(middlewares=[sm])

    with pytest.raises(
        ValidationError,
        match="Validation error by middleware _SimpleMiddleware, at index 0.",
    ):
        pipe.validate(set(), set(), sized_data_loader=False)


def test_validate_ok() -> None:
    sm1 = _SimpleMiddleware()
    sm1.requires_state_fields = {"miss"}
    sm1.provides_data_fields_pre = {"val"}
    sm1.requires_data_fields_post = {"val", "val1"}

    sm2 = _SimpleMiddleware()
    sm2.requires_data_fields_pre = {"val"}
    sm2.provides_data_fields_post = {"val1"}

    pipe = Pipeline(middlewares=[sm1, sm2])

    pipe.validate({"miss"}, set(), sized_data_loader=False)


def test_validate_ok_sized() -> None:
    sm1 = _SimpleMiddleware()
    sm1.requires_state_fields = {"miss", "num_steps"}
    sm1.provides_data_fields_pre = {"val"}
    sm1.requires_data_fields_post = {"val", "val1"}

    sm2 = _SimpleMiddleware()
    sm2.requires_data_fields_pre = {"val"}
    sm2.provides_data_fields_post = {"val1"}

    pipe = Pipeline(middlewares=[sm1, sm2])

    pipe.validate({"miss"}, set(), sized_data_loader=True)


def test_run_missing_state_field() -> None:
    sm = _SimpleMiddleware()
    sm.requires_state_fields = {"miss"}

    pipe = Pipeline(middlewares=[sm])

    with pytest.raises(
        ValidationError,
        match="Validation error by middleware _SimpleMiddleware, at index 0.",
    ):
        pipe.run({}, EmptyLoader(3))


def test_run_missing_data_field() -> None:
    sm = _SimpleMiddleware()
    sm.requires_data_fields_pre = {"miss"}

    pipe = Pipeline(middlewares=[sm])

    with pytest.raises(
        ValidationError,
        match="Validation error by middleware _SimpleMiddleware, at index 0.",
    ):
        pipe.run({}, EmptyLoader(3))


def test_run_missing_data_field_no_validate() -> None:
    """
    Check that with validation turned off the pipeline will run despite a
    missing data field.
    """
    sm = _SimpleMiddleware()
    sm.requires_data_fields_pre = {"miss"}

    pipe = Pipeline(middlewares=[sm])

    pipe.run({}, EmptyLoader(3), validate=False)


def test_run_ok() -> None:
    class _Middleware(Middleware[Any, Any]):
        def wrap(self, next_step: ProcessingStep[Any, Any]) -> ProcessingStep[Any, Any]:
            def wrapped(state: Any, data: Any) -> None:
                state["value"] = state["step"] + 1

            return wrapped

    state: Any = {"acc": []}

    pipeline = Pipeline(middlewares=[AccMiddleware(), _Middleware()])
    pipeline.run(state=state, data_loader=EmptyLoader(3))

    assert state["acc"] == [1, 2, 3]


def test_run_no_loader_length() -> None:
    class _Middleware(Middleware[Any, Any]):
        def wrap(self, next_step: ProcessingStep[Any, Any]) -> ProcessingStep[Any, Any]:
            def wrapped(state: Any, data: Any) -> None:
                assert set(state.keys()) == {"step"}

            return wrapped

    pipeline = Pipeline(middlewares=[_Middleware()])

    data_loader: WrappedUnsizedLoader = wrap_iterable(([] for _ in range(3)), ())
    state: Any = {}

    pipeline.run(state, data_loader)


def test_run_dataloader_length() -> None:
    data_length = 3

    class _Middleware(Middleware[Any, Any]):
        def wrap(self, next_step: ProcessingStep[Any, Any]) -> ProcessingStep[Any, Any]:
            def wrapped(state: Any, data: Any) -> None:
                assert set(state.keys()) == {"epoch", "num_epochs"}
                assert state["num_epochs"] == data_length

            return wrapped

    pipeline = Pipeline(middlewares=[_Middleware()], step_name="epoch")
    state: Any = {}

    pipeline.run(state, EmptyLoader(data_length))


def test_skip_step() -> None:
    class _Middleware(Middleware[Any, Any]):
        def wrap(self, next_step: ProcessingStep[Any, Any]) -> ProcessingStep[Any, Any]:
            def wrapped(state: Any, data: Any) -> None:
                state["value"] = state["step"] + 1
                if state["step"] == 1:
                    raise SkipStep

            return wrapped

    state: Any = {"acc": []}

    pipeline = Pipeline(middlewares=[AccMiddleware(), _Middleware()])
    pipeline.run(state=state, data_loader=EmptyLoader(3))

    assert state["acc"] == [1, 3]


def test_abort_pipeline() -> None:
    class _Middleware(Middleware[Any, Any]):
        def wrap(self, next_step: ProcessingStep[Any, Any]) -> ProcessingStep[Any, Any]:
            def wrapped(state: Any, data: Any) -> None:
                state["value"] = state["step"] + 1
                if state["step"] == 1:
                    raise AbortPipeline

            return wrapped

    state: Any = {"acc": []}

    pipeline = Pipeline(middlewares=[AccMiddleware(), _Middleware()])
    pipeline.run(state=state, data_loader=EmptyLoader(3))

    assert state["acc"] == [1]


def test_middleware_callbacks() -> None:
    mware = _SimpleMiddleware()

    mware.on_start = MagicMock(wraps=mware.on_start)  # type: ignore[method-assign]
    mware.on_finish = MagicMock(wraps=mware.on_finish)  # type: ignore[method-assign]

    state: Any = {"hello": 1}

    pipeline = Pipeline(middlewares=[mware])
    pipeline.run(state=state, data_loader=EmptyLoader(3))

    # This test is not exactly correct - because the state is mutable, and changes in
    # the course of execution. so the final state will have "step" and "value" keys,
    # which the initial state (that was passed to the on_start call) did not have.
    # Nevertheless, as MagicMock simply stores the reference to the arguments passed
    # in the call (and does not copy them), this check passes.
    mware.on_start.assert_called_once_with(state)

    mware.on_finish.assert_called_once_with(state)


def test_pipeline_wrapper_ok() -> None:
    mware = _SimpleMiddleware()
    mware.requires_state_fields = {"epoch", "num_epochs", "num_steps", "step"}

    inner_pipeline = Pipeline([mware])

    pipe_mware = PipelineWrapper(inner_pipeline, EmptyLoader(5))

    outer_pipeline = Pipeline([pipe_mware], step_name="epoch")

    state: dict[str, Any] = {}
    outer_pipeline.run(state, EmptyLoader(3))

    # State after all pipelines/loops have been run
    assert state == {"epoch": 2, "num_epochs": 3, "num_steps": 5, "step": 4}



"""
Module defining data-loading classes and protocols.

This module exports Loader, which is a common protocol for data loaders consumed by
pipelines, and also provides EmptyLoader, a range equivalent for data loaders, and
wrap_iterable, which wraps an iterable that returns tuples as batches, so that it
conforms to the Loader protocol (returns dicts as batches).
"""

from collections.abc import Iterable, Iterator, Sequence, Sized
from typing import Any, Protocol, cast

from .middleware import StrMapping

__all__ = [
    "EmptyLoader",
    "Loader",
    "WrappedSizedLoader",
    "WrappedUnsizedLoader",
    "wrap_iterable",
]


class Loader(Protocol, Iterable[StrMapping]):
    """
    A protocol for all data loaders consumed by pipelines.

    A loader should be an iterable that yields string-key mappings as batches
    of data, and should also have a `data_fields` attribute, which is used for
    validating the pipeline.
    """

    data_fields: set[str]


class EmptyLoader:
    """
    A loader that returns empty batces of data.

    Useful for iteatring over epochs.

    Example usage:
        >>> generator = EmptyGenerator(num_steps=3)
        >>> for batch in generator:
        ...     print(batch)
        {}
        {}
        {}
        >>> print(len(generator))
        3
    """

    def __init__(self, num_steps: int) -> None:
        """
        Initialize an EmptyGenerator instance.

        Args:
            num_steps: The number of steps (or batches) to generate.

        """
        self.num_steps = num_steps
        self.data_fields: set[str] = set()

    def __iter__(self) -> Iterator[StrMapping]:
        """
        Create an iterator that yields empty dictionaries for each step.

        Yields:
            An iterator yielding empty dictionaries.

        """
        for _ in range(self.num_steps):
            yield {}

    def __len__(self) -> int:
        """
        Return the total number of steps.

        Returns:
            The number of steps.

        """
        return self.num_steps


class WrappedUnsizedLoader:
    """
    A wrapper for iterable data loaders that return batches as sequences.

    This class converts each sequence-based batch from the wrapped loader into
    a dictionary using the fields specified in `data_fields`.
    """

    def __init__(
        self, loader: Iterable[Sequence[Any]], data_fields: Sequence[str]
    ) -> None:
        """
        Initialize a WrappedUnsizedLoader.

        Args:
            loader: An iterable that yields sequences (e.g., tuples or lists)
                representing data batches.
            data_fields: A sequence of string keys to use in the output dictionaries.
                The number of keys must match the length of each sequence in the loader,
                each item in batch will be assigned the key at the corresponding
                position.

        """
        self._loader = loader
        self.data_fields = set(data_fields)
        self.data_fields_seq = data_fields

    def __iter__(self) -> Iterator[StrMapping]:
        """
        Create an iterator that converts each sequence batch into a dictionary.

        Yields:
            A dict mapping each item of the sequence to the corresponding key
            in `data_fields`.

        """
        for batch in self._loader:
            batch_dict: dict[str, Any] = dict(
                zip(self.data_fields_seq, batch, strict=True)
            )
            yield batch_dict


class WrappedSizedLoader(WrappedUnsizedLoader):
    """
    A wrapper for iterable data loaders that return sequences and have a length.

    In addition to converting sequence-based batches into dictionaries,
    this class implements the __len__ method, allowing callers to check
    the total number of batches.
    """

    def __len__(self) -> int:
        """
        Return the number of batches in the wrapped loader.

        Returns:
            The length of the underlying loader.

        """
        return len(cast(Sized, self._loader))


def wrap_iterable(
    loader: Iterable[Sequence[Any]], data_fields: Sequence[str]
) -> WrappedSizedLoader | WrappedUnsizedLoader:
    """
    Wrap a sequence-yielding iterable, so that it conforms to Loader protocol.

    This will wrap an iterable, which yields a sequence (tuple, list, ...) as a batch
    in a wrapper class, which simply puts each item of the original data loader batch
    in a dict, with keys specified in `data_fields` argument.

    If the original data loader is a sizable (has length), the wrapped object
    will also be.

    Arguments:
        loader: An iterable, which yields sequences (tuples, lists, ...). Can be sized
            (has length) or not.
        data_fields: Keys for data fields in the dict of the batches of the data loader
            wrapper. The number of keys has to match the number items in a batch of
            `loader`, each item in batch will be assigned the key at the corresponding
            position.

    Returns:
        A wrapped data loader, that yields dictionary batches.

    """
    if isinstance(loader, Sized):
        return WrappedSizedLoader(loader, data_fields)

    return WrappedUnsizedLoader(loader, data_fields)
